#include <stdint.h>
#include "MKL26Z4.h"

#include <dm_1wire_control.h>

/*
Pins used in this application:

PTC01 [O] A5	- TPM0_CH0: output falling edge indicator (debugging)
PTC02 [O] A4	- TPM0_CH1: output rising edge indicator (debugging)
PTD02 [I] D9	- TPM0_CH2: falling edge capture input (optional)
PTD04 [I] D10	- TPM0_CH4: rising edege capture input (MUST HAVE!!!)

PTB03 [X] A3	- GPIO		: 1-wire bus (created by using output and input port setup via DMA transfers)

PTB00	[O] A0	- GLOBAL	: indicator line (debugging)
PTB01	[O] A1	- TEST		: indicator line (debugging)

*/

#define	TPM_CH_OUT_FE			0
#define	TPM_CH_OUT_RE			1
#define	TPM_CH_IN_FE			2
#define	TPM_CH_IN_RE			4

#define	DM1WIREBUS_PORT		PORTB
#define	DM1WIREBUS_IOPORT	PTB
#define	DM1WIREBUS_PIN		3

#define	GLOBAL_PORT				PORTB
#define	GLOBAL_IOPORT			PTB
#define	GLOBAL_PIN				0

#define	TEST_PORT					PORTB
#define	TEST_IOPORT				PTB
#define	TEST_PIN					1

//the following setup divides the TPM clock by 2; since it is the FLL = 42 MHz driving TPM0, 
//after prescaling by 2 it takes 21 counts to generate 1 us
//the goal is to keep the 1 ms count under 64K so adjust the prescale as needed
#define	TPM_PRESCALE			1	
#define	COUNT_PER_US			21
#define	DM1W_OFFSET				(  1*COUNT_PER_US)	//activities start with this delay to timer = 0
#define	DM1W_WR0LOW				( 60*COUNT_PER_US)	//bus stays this low when master writes 0
#define	DM1W_WR1LOW				(  2*COUNT_PER_US)	//bus stays this low when master writes 1
#define	DM1W_RDLOW				(  2*COUNT_PER_US)	//bus stays this low when master requests (reads) data from slave
#define	DM1W_1BITPERIOD		( 65*COUNT_PER_US)	//overall 1 bit takes this long to be written/read
#define	DM1W_RSTLOW				(480*COUNT_PER_US)	//reset-presence activity low state
#define	DM1W_RST_PERIOD		(960*COUNT_PER_US)	//reset-presence activity overall duration

#define	DM1W_RD_THRESHOLD	(13*COUNT_PER_US)		//master reading data from slave and seeing low pulse shorter than this count declars bit=1; otherwise bit=0

volatile uint32_t drive_1wbus_low, drive_1wbus_high;
volatile uint32_t timestamp_fe,	timestamp_re;

//this is basic 1-wire initialization
void dm_1wire_init(void)
{
	uint32_t dummy_loc;
	
	//enable access to all ports
	SIM->SCGC5 |= SIM_SCGC5_PORTE_MASK | 
								SIM_SCGC5_PORTD_MASK |	
								SIM_SCGC5_PORTC_MASK |	
								SIM_SCGC5_PORTB_MASK |	
								SIM_SCGC5_PORTA_MASK;
	
	//GLOBAL: output, low
	GLOBAL_IOPORT->PDDR |= 1<<GLOBAL_PIN;
	GLOBAL_IOPORT->PCOR = 1<<GLOBAL_PIN;
	GLOBAL_PORT->PCR[GLOBAL_PIN] = PORT_PCR_MUX(1);

	//TEST: output, low
	TEST_IOPORT->PDDR |= 1<<TEST_PIN;
	TEST_IOPORT->PCOR = 1<<TEST_PIN;
	TEST_PORT->PCR[TEST_PIN] = PORT_PCR_MUX(1);
	
	//1-wire bus: input yet drives low when output
	DM1WIREBUS_IOPORT->PDDR &= ~(1<<DM1WIREBUS_PIN);
	DM1WIREBUS_IOPORT->PCOR = 1<<DM1WIREBUS_PIN;
	DM1WIREBUS_PORT->PCR[DM1WIREBUS_PIN] = PORT_PCR_MUX(1);
	
	drive_1wbus_low = DM1WIREBUS_IOPORT->PDDR | (1<<DM1WIREBUS_PIN);
	drive_1wbus_high= DM1WIREBUS_IOPORT->PDDR & ~(1<<DM1WIREBUS_PIN);
	
	PORTC->PCR[1] = PORT_PCR_MUX(4);		//PTC1: TPM0_CH0 - debugging
	PORTC->PCR[2] = PORT_PCR_MUX(4);		//PTC2: TPM0_CH1 - debugging
	PORTD->PCR[2] = PORT_PCR_MUX(4);		//PTD2: TPM0_CH2 - optional 
	PORTD->PCR[4] = PORT_PCR_MUX(4);		//PTD4: TPM0_CH4 - must have
	
	//
	// TPM0 setup begin
	//==================
	
	//TPM clock: FLL
	SIM->SOPT2 = (SIM->SOPT2 & ~(SIM_SOPT2_TPMSRC_MASK | SIM_SOPT2_PLLFLLSEL_MASK)) | 
															((SIM_SOPT2_TPMSRC(1)) | (0<<SIM_SOPT2_PLLFLLSEL_SHIFT));
	
	SIM->SCGC6 |= SIM_SCGC6_TPM0_MASK;	//enable TPM0 clock
	
	TPM0->SC =	0<<TPM_SC_DMA_SHIFT 	|	//no DMA
							0<<TPM_SC_TOF_SHIFT		|	//NA
							1<<TPM_SC_TOIE_SHIFT	|	//enable TOF int
							0<<TPM_SC_CPWMS_SHIFT	|	//up counting
							TPM_SC_CMOD(0)				|	//timer disabled
							TPM_SC_PS(TPM_PRESCALE);//as per setup
							
	TPM0->CNT = 0x1234;									//write anything to reset
	TPM0->MOD = DM1W_OFFSET+DM1W_RST_PERIOD-1;	//assume reset and presence pulse period

	//output falling edge channel (TPM_CH_OUT_FE)
	//MS[B:A]=10,ELS[BA]=10 => Edge alligned PWM: High-true pulses (clear Output on match, set Output on reload)
	TPM0->CONTROLS[TPM_CH_OUT_FE].CnSC = 
							0<<TPM_CnSC_CHIE_SHIFT	|	//NA
							0<<TPM_CnSC_CHIE_SHIFT	|	//no interrupt
							1<<TPM_CnSC_MSB_SHIFT 	|	//MSB=1
							0<<TPM_CnSC_MSA_SHIFT 	|	//MSA=0
							1<<TPM_CnSC_ELSB_SHIFT 	|	//ELSB=1
							0<<TPM_CnSC_ELSA_SHIFT 	|	//ELSA=0
							1<<TPM_CnSC_DMA_SHIFT;		//enable DMA request
	TPM0->CONTROLS[TPM_CH_OUT_FE].CnV = DM1W_OFFSET;

	//output rising edge channel (TPM_CH_OUT_RE)
	//MS[B:A]=10,ELS[BA]=10 => Edge alligned PWM: High-true pulses (clear Output on match, set Output on reload)
	TPM0->CONTROLS[TPM_CH_OUT_RE].CnSC = 
							0<<TPM_CnSC_CHIE_SHIFT	|	//NA
							0<<TPM_CnSC_CHIE_SHIFT	|	//no interrupt
							1<<TPM_CnSC_MSB_SHIFT 	|	//MSB=1
							0<<TPM_CnSC_MSA_SHIFT 	|	//MSA=0
							1<<TPM_CnSC_ELSB_SHIFT 	|	//ELSB=1
							0<<TPM_CnSC_ELSA_SHIFT 	|	//ELSA=0
							1<<TPM_CnSC_DMA_SHIFT;		//enable DMA request
	TPM0->CONTROLS[TPM_CH_OUT_RE].CnV = DM1W_OFFSET+DM1W_RSTLOW;
	
	//input falling edge only channel (TPM_CH_IN_FE)
	//MS[B:A]=00,ELS[BA]=10 => Input capture (falling edge only)
	TPM0->CONTROLS[TPM_CH_IN_FE].CnSC = 
							0<<TPM_CnSC_CHIE_SHIFT	|	//NA
							0<<TPM_CnSC_CHIE_SHIFT	|	//no interrupt
							0<<TPM_CnSC_MSB_SHIFT 	|	//MSB=0
							0<<TPM_CnSC_MSA_SHIFT 	|	//MSA=0
							1<<TPM_CnSC_ELSB_SHIFT 	|	//ELSB=1
							0<<TPM_CnSC_ELSA_SHIFT 	|	//ELSA=0
							0<<TPM_CnSC_DMA_SHIFT;		//no DMA request
	dummy_loc = TPM0->CONTROLS[TPM_CH_IN_FE].CnV;
	
	//input rising edge only channel (TPM_CH_IN_RE)
	//MS[B:A]=00,ELS[BA]=01 => Input capture (rising edge only)
	TPM0->CONTROLS[TPM_CH_IN_RE].CnSC = 
							0<<TPM_CnSC_CHIE_SHIFT	|	//NA
							0<<TPM_CnSC_CHIE_SHIFT	|	//no interrupt
							0<<TPM_CnSC_MSB_SHIFT 	|	//MSB=0
							0<<TPM_CnSC_MSA_SHIFT 	|	//MSA=0
							0<<TPM_CnSC_ELSB_SHIFT 	|	//ELSB=1
							1<<TPM_CnSC_ELSA_SHIFT 	|	//ELSA=0
							0<<TPM_CnSC_DMA_SHIFT;		//no DMA request
	dummy_loc = TPM0->CONTROLS[TPM_CH_IN_RE].CnV;

	TPM0->STATUS = 	TPM_STATUS_TOF_MASK | 
									1<<TPM_CH_OUT_FE		|
									1<<TPM_CH_OUT_RE		|
									1<<TPM_CH_IN_FE			|
									1<<TPM_CH_IN_RE;
									
	TPM0->CONF = 	TPM_CONF_TRGSEL(0)				|	//NA
								0<<TPM_CONF_CROT_SHIFT		|	//NA
								1<<TPM_CONF_CSOO_SHIFT		| //counter stops on overflow
								0<<TPM_CONF_CSOT_SHIFT		| //counter starts incrementing immediatelly once enabled
								0<<TPM_CONF_GTBEEN_SHIFT	|	//all channels use the internally generated TPM as timebase
								TPM_CONF_DBGMODE(0)				|	//timer stops in debug mode
								0<<TPM_CONF_DOZEEN_SHIFT;		//timer continues in wait mode
	
	// TPM0 setup end
	//----------------
	
	//
	// basic DMA setup begin
	//=======================

	SIM->SCGC7 |= SIM_SCGC7_DMA_MASK;					//enable DMA clock
	SIM->SCGC6 |= SIM_SCGC6_DMAMUX_MASK;			//enable DMAMUX clock
	
	//DMACH0: generate falling edge (TPM0_CH0)
	DMAMUX0->CHCFG[0] = DMAMUX_CHCFG_ENBL_MASK | DMAMUX_CHCFG_SOURCE(24);
	//source: drive 1wire bus low pattern
	DMA0->DMA[0].SAR = (uint32_t) &drive_1wbus_low;
	//destination: port register in charge of driving 1wire bus low
	DMA0->DMA[0].DAR = (uint32_t) &DM1WIREBUS_IOPORT->PDDR;
	
	//DMACH1: generate rising edge (TPM0_CH1)
	DMAMUX0->CHCFG[1] = DMAMUX_CHCFG_ENBL_MASK | DMAMUX_CHCFG_SOURCE(25);
	//source: drive 1wire bus high pattern
	DMA0->DMA[1].SAR = (uint32_t) &drive_1wbus_high;
	//destination: port register in charge of driving 1wire bus high
	DMA0->DMA[1].DAR = (uint32_t) &DM1WIREBUS_IOPORT->PDDR;

	// basic DMA setup end
	//---------------------
	
	return;
}

//generate the reset pulse (1-wire bus low for 480 us); after 960 us return
//delay all activities by the offset (inserts a period of idle bus no matter what)
//use test pin to indicate overall routine length
//
//there will be only 1 DMA CH0 transfer (only 1 falling edge) => 4 bytes will be sent; similar for CH1 and rising edge
//
void dm_1wire_reset_pulse(void)
{
	//clear flags
	TPM0->STATUS = 	TPM_STATUS_TOF_MASK | 
									1<<TPM_CH_OUT_FE		|
									1<<TPM_CH_OUT_RE		|
									1<<TPM_CH_IN_FE			|
									1<<TPM_CH_IN_RE;
	
	TPM0->SC =	0<<TPM_SC_DMA_SHIFT 	|	//no DMA
							0<<TPM_SC_TOF_SHIFT		|	//NA
							1<<TPM_SC_TOIE_SHIFT	|	//enable TOF int
							0<<TPM_SC_CPWMS_SHIFT	|	//up counting
							TPM_SC_CMOD(0)				|	//disable timer
							TPM_SC_PS(TPM_PRESCALE);//as per setup
	
	TPM0->CNT = 0x1234;																						//write anything to reset the timer
	TPM0->MOD = DM1W_OFFSET+DM1W_RST_PERIOD-1;										//prepare for reset and presence pulse period
	TPM0->CONTROLS[TPM_CH_OUT_FE].CnV = DM1W_OFFSET;							//preapre for falling edge
	TPM0->CONTROLS[TPM_CH_OUT_RE].CnV = DM1W_OFFSET+DM1W_RSTLOW;	//prepare for rising edge
	
	//clear DONE bit;prepare to transfer 1 byte
	DMA0->DMA[0].DSR_BCR = DMA_DSR_BCR_DONE_MASK;
	DMA0->DMA[0].DSR_BCR = 1*4;
	//no EOT int,enable peri request,single transfer per request, disable auto-align,
	//per RM,disable async req,do not inc source,32 bit source data,
	//do not inc destination,32 bit destination data,do not start,source buffer disabled
	//destination buffer disabled,clear ERQ at end,no ch linking
	DMA0->DMA[0].DCR = 	0UL<<DMA_DCR_EINT_SHIFT	| 1<<DMA_DCR_ERQ_SHIFT		| 1<<DMA_DCR_CS_SHIFT			| 0<<DMA_DCR_AA_SHIFT	|\
											0<<24										| 0<<DMA_DCR_EADREQ_SHIFT	| 0<<DMA_DCR_SINC_SHIFT		| DMA_DCR_SSIZE(0)		|\
											0<<DMA_DCR_DINC_SHIFT		| DMA_DCR_DSIZE(0)				| 0<<DMA_DCR_START_SHIFT	| DMA_DCR_SMOD(0)			|\
											DMA_DCR_DMOD(0)					| 1<<DMA_DCR_D_REQ_SHIFT	| DMA_DCR_LINKCC(0);
	
	//clear DONE bit;prepare to transfer 1 byte
	DMA0->DMA[1].DSR_BCR = DMA_DSR_BCR_DONE_MASK;
	DMA0->DMA[1].DSR_BCR = 1*4;
	//no EOT int,enable peri request,single transfer per request, disable auto-align,
	//per RM,disable async req,do not inc source,32 bit source data,
	//do not inc destination,32 bit destination data,do not start,source buffer disabled
	//destination buffer disabled,clear ERQ at end,no ch linking
	DMA0->DMA[1].DCR = 	0UL<<DMA_DCR_EINT_SHIFT	| 1<<DMA_DCR_ERQ_SHIFT		| 1<<DMA_DCR_CS_SHIFT			| 0<<DMA_DCR_AA_SHIFT	|\
											0<<24										| 0<<DMA_DCR_EADREQ_SHIFT	| 0<<DMA_DCR_SINC_SHIFT		| DMA_DCR_SSIZE(0)		|\
											0<<DMA_DCR_DINC_SHIFT		| DMA_DCR_DSIZE(0)				| 0<<DMA_DCR_START_SHIFT	| DMA_DCR_SMOD(0)			|\
											DMA_DCR_DMOD(0)					| 1<<DMA_DCR_D_REQ_SHIFT	| DMA_DCR_LINKCC(0);
	
	TEST_IOPORT->PCOR = 1<<TEST_PIN;		//indicate reset and presence pulse period start
	
	TPM0->SC =	0<<TPM_SC_DMA_SHIFT 	|	//no DMA
							0<<TPM_SC_TOF_SHIFT		|	//NA
							1<<TPM_SC_TOIE_SHIFT	|	//enable TOF int
							0<<TPM_SC_CPWMS_SHIFT	|	//up counting
							TPM_SC_CMOD(1)				|	//timer increments on every TPM counter clock
							TPM_SC_PS(TPM_PRESCALE);//as per setup
	
	while((TPM0->STATUS & TPM_STATUS_TOF_MASK) == 0);	//wait for the timer to stop
	
	//clear flags => cancel DMA requests
	TPM0->STATUS = 	TPM_STATUS_TOF_MASK | 
									1<<TPM_CH_OUT_FE		|
									1<<TPM_CH_OUT_RE		|
									1<<TPM_CH_IN_FE			|
									1<<TPM_CH_IN_RE;
	
	TEST_IOPORT->PSOR = 1<<TEST_PIN;		//indicate reset and presence pulse period end
	
	return;
}

//generate 1 bit waveform on the 1-wire bus
//TPM_CH_OUT_FE and TPM_CH_OUT_RE are used to drive DMA requests; DMA requests toggle the 1-wire dedicated pin
//from an output pin (driving low) to an input pin (letting the slave and external pull-up control the bus)
//once the timer reaches its period it stops
//TPM_CH_IN_FE and TPM_CH_IN_RE are falling and rising edge sensitive timer channels; by subtracting the 
//falling edge captured timestamp from the rising edge timestamp one can measure the width of the negative
//pulse on the 1-wire bus; when reading data from the slave compare this difference against the preset
//threshold (13-15 us) and bit content can be figured out
//in order to read a byte from the slave send 0xFF
uint32_t dm_1wire_txrx_1bit(uint32_t tx_bit)
{
	uint32_t rx_bit;
	
	TPM0->STATUS = 	TPM_STATUS_TOF_MASK | 
									1<<TPM_CH_OUT_FE		|
									1<<TPM_CH_OUT_RE		|
									1<<TPM_CH_IN_FE			|
									1<<TPM_CH_IN_RE;
	
	TPM0->SC =	0<<TPM_SC_DMA_SHIFT 	|	//no DMA
							0<<TPM_SC_TOF_SHIFT		|	//NA
							1<<TPM_SC_TOIE_SHIFT	|	//enable TOF int
							0<<TPM_SC_CPWMS_SHIFT	|	//up counting
							TPM_SC_CMOD(0)				|	//disable timer
							TPM_SC_PS(TPM_PRESCALE);//as per setup
	
	TPM0->CNT = 0x1234;																						//write anything to reset
	TPM0->MOD = DM1W_OFFSET+DM1W_1BITPERIOD-1;										//prepare to send/receive 1 bit of data
	TPM0->CONTROLS[TPM_CH_OUT_FE].CnV = DM1W_OFFSET;							//preapre for falling edge
	if (tx_bit == 0)																							//prepare for rising edge
	{
		TPM0->CONTROLS[TPM_CH_OUT_RE].CnV = DM1W_OFFSET+DM1W_WR0LOW;//prepare rising edge matching bit=0
	}
	else
	{
		TPM0->CONTROLS[TPM_CH_OUT_RE].CnV = DM1W_OFFSET+DM1W_WR1LOW;//prepare rising edge matching bit=1
	}
	
	TPM0->SC =	0<<TPM_SC_DMA_SHIFT 	|	//no DMA
							0<<TPM_SC_TOF_SHIFT		|	//NA
							1<<TPM_SC_TOIE_SHIFT	|	//enable TOF int
							0<<TPM_SC_CPWMS_SHIFT	|	//up counting
							TPM_SC_CMOD(1)				|	//timer increments on every TPM counter clock
							TPM_SC_PS(TPM_PRESCALE);//as per setup
	
	while((TPM0->STATUS & TPM_STATUS_TOF_MASK) == 0);	//wait for the timer to stop/reach the preset period
	
	timestamp_fe = TPM0->CONTROLS[TPM_CH_IN_FE].CnV;	//extract falling edge timestamp
	timestamp_re = TPM0->CONTROLS[TPM_CH_IN_RE].CnV;	//extract rising edge timestamp
	
	//comapre timestamps delta with the threshold and extract bit content
	if ((timestamp_re - timestamp_fe) > DM1W_RD_THRESHOLD)
	{
		rx_bit = 0;
	}
	else
	{
		rx_bit = 1;
	}
	
	return rx_bit;
}

//use this routine to transmit and receive a number of bytes over the 1-wire bus
//
//number of bytes to transfer per DMA channel = no_of_bytes*8*4
// - 1 32-bit register (4 byute) write per falling/rising edge;
// - 8 bits per byte of data
// - no_of_bytes set by the user
uint32_t dm_1wire_txrx(uint8_t* p_tx_data, uint8_t* p_rx_data, uint32_t txrx_data_cnt)
{
	uint32_t result;
	uint32_t i_loc, bit_index;
	uint8_t temp_rx_data;
	
	if (0 == txrx_data_cnt)
	{
		result = 1;
	}
	else
	{
		result = 0;

		//clear DONE bit;prepare bytes for as many bits as needed
		DMA0->DMA[0].DSR_BCR = DMA_DSR_BCR_DONE_MASK;
		DMA0->DMA[0].DSR_BCR = txrx_data_cnt*8*4;
		//no EOT int,enable peri request,single transfer per request, disable auto-align,
		//per RM,disable async req,do not inc source,32 bit source data,
		//do not inc destination,32 bit destination data,do not start,source buffer disabled
		//destination buffer disabled,clear ERQ at end,no ch linking
		DMA0->DMA[0].DCR = 	0UL<<DMA_DCR_EINT_SHIFT	| 1<<DMA_DCR_ERQ_SHIFT		| 1<<DMA_DCR_CS_SHIFT			| 0<<DMA_DCR_AA_SHIFT	|\
												0<<24										| 0<<DMA_DCR_EADREQ_SHIFT	| 0<<DMA_DCR_SINC_SHIFT		| DMA_DCR_SSIZE(0)		|\
												0<<DMA_DCR_DINC_SHIFT		| DMA_DCR_DSIZE(0)				| 0<<DMA_DCR_START_SHIFT	| DMA_DCR_SMOD(0)			|\
												DMA_DCR_DMOD(0)					| 1<<DMA_DCR_D_REQ_SHIFT	| DMA_DCR_LINKCC(0);

		//clear DONE bit;prepare bytes for as many bits as needed
		DMA0->DMA[1].DSR_BCR = DMA_DSR_BCR_DONE_MASK;
		DMA0->DMA[1].DSR_BCR = txrx_data_cnt*8*4;
		//no EOT int,enable peri request,single transfer per request, disable auto-align,
		//per RM,disable async req,do not inc source,32 bit source data,
		//do not inc destination,32 bit destination data,do not start,source buffer disabled
		//destination buffer disabled,clear ERQ at end,no ch linking
		DMA0->DMA[1].DCR = 	0UL<<DMA_DCR_EINT_SHIFT	| 1<<DMA_DCR_ERQ_SHIFT		| 1<<DMA_DCR_CS_SHIFT			| 0<<DMA_DCR_AA_SHIFT	|\
												0<<24										| 0<<DMA_DCR_EADREQ_SHIFT	| 0<<DMA_DCR_SINC_SHIFT		| DMA_DCR_SSIZE(0)		|\
												0<<DMA_DCR_DINC_SHIFT		| DMA_DCR_DSIZE(0)				| 0<<DMA_DCR_START_SHIFT	| DMA_DCR_SMOD(0)			|\
												DMA_DCR_DMOD(0)					| 1<<DMA_DCR_D_REQ_SHIFT	| DMA_DCR_LINKCC(0);
		
		
		for (i_loc = 0; i_loc != txrx_data_cnt; i_loc++)
		{
			temp_rx_data = 0;
			for (bit_index = 0; bit_index != 8; bit_index++)
			{
				temp_rx_data |= ((uint8_t)dm_1wire_txrx_1bit((p_tx_data[i_loc]>>bit_index) & 0x01))<<bit_index;
				}//end of loop: bits within the byte
			p_rx_data[i_loc] = temp_rx_data;
			}//end of loop: user data
		}//end of: non-0 transfer data count
	
	return result;
}
